"""
search.py

@author: DevelopmentSeed

Backend algorithms for finding tiles within a geospatial boundary
"""
from queue import LifoQueue

from shapely.geometry import shape
from pygeotile.tile import Tile

from geodex.utils import get_overlap_child_tiles, format_tile


def tile_search(boundary, max_zoom, tile_format='google',
                output_format='{x} {y} {z}', print_to=None):
    """Print (or save) tiles belonging to a shape object.

    Parameters:
    ----------
    boundary: str
        JSON corresponding to a single polygon/multipolygon
    max_zoom: int
        Zoom at which to terminate file search
    tile_format: str
        Desired tile format. `google` (default), `tms`, or `quad_tree`
    output_format: str
        Format string for printing the tile indices. Only used for tile_format
        `google` and `tms`. Must contain `{x}`, `{y}`, and `{z}`.
    print_to: object with `write` functionality
        `None` (default) prints to sys.stdout. Pass a (non-binary) file object
        to print directly to an open file.

    Returns:
    -------
    total_tiles: int
        Number of tiles found

    Note: this function is useful if you want to directly incorporate tile
    searching into a script (i.e., not on the command line).
    """
    #########################################
    # Initialize boundary and queue
    #########################################
    bound_shape = shape(boundary)  # Convert JSON to shape object
    stack = LifoQueue()
    total_tiles = 0

    # Put whole-world tile on stack
    start_tile = Tile.from_google(0, 0, 0)
    stack.put([start_tile, False])

    #########################################
    # Depth-first search on tile indices
    #########################################
    while not stack.empty():

        # Pop the top tile in the stack
        top_tile, comp_contained = stack.get()

        # Check if desired zoom has been reached
        if top_tile.zoom >= max_zoom:
            # If at max zoom, print tile inds or save to file (if specified)
            print(format_tile(top_tile, tile_format, output_format),
                  file=print_to)
            total_tiles += 1

        # Otherwise, zoom in one increment, find children tiles, add to stack
        else:
            ret_tiles = get_overlap_child_tiles(top_tile, bound_shape,
                                                comp_contained)
            for rt in ret_tiles:
                stack.put(rt)

    return total_tiles


def process_geojson(boundaries, max_zoom, tile_format, output_format):
    """Process entire list of boundaries."""
    total_tiles = 0

    # Loop through all boundaries and compute tiles for each
    for bound in boundaries:
        bound_tiles = tile_search(bound, max_zoom, tile_format, output_format)
        total_tiles += bound_tiles

    return total_tiles



"""
test_utils.py

@author: DevelopmentSeed

Test all utilities for GeoDex.
"""
import os
import unittest
from unittest.mock import patch
from io import StringIO

from geodex.search import tile_search, process_geojson
from geodex.utils import load_geojson

testpath = os.path.dirname(__file__)

class TestCore(unittest.TestCase):
    """Test core functionality."""

    @patch('sys.stdout', new_callable=StringIO)
    def test_single_boundary_tile_search(self, mock_stdout):

        """Test that tiles are correctly found for a single boundary."""
        tile_inds = ('2494 2126 12\n'
                     '2494 2125 12\n'
                     '2493 2126 12\n'
                     '2493 2125 12\n')
        tile_inds_tms_zxy = ('12-2494-1969\n'
                             '12-2494-1970\n'
                             '12-2493-1969\n'
                             '12-2493-1970\n')
        tile_inds_quad_tree = ('300112113330\n'
                               '300112113312\n'
                               '300112113321\n'
                               '300112113303\n')

        # Load geojson, get 0th object
        shape_json = load_geojson('./roi_single.geojson')[0]

        num_tiles = tile_search(shape_json, max_zoom=12)
        self.assertEqual(num_tiles, 4)
        self.assertEqual(mock_stdout.getvalue(), tile_inds)
        mock_stdout.seek(0)  # XXX: hacky way to reset StringIO object
        mock_stdout.truncate(0)

        num_tiles = tile_search(shape_json, 12, 'tms', '{z}-{x}-{y}')
        self.assertEqual(mock_stdout.getvalue(), tile_inds_tms_zxy)
        mock_stdout.seek(0) # XXX: hacky way to reset StringIO object
        mock_stdout.truncate(0)

        num_tiles = tile_search(shape_json, 12, 'quad_tree')
        self.assertEqual(mock_stdout.getvalue(), tile_inds_quad_tree)


    @patch('sys.stdout', new_callable=StringIO)
    def test_multiboundary_tile_search(self, mock_stdout):
        """Test that tiles are correctly found for multiple boundaries."""
        tile_inds = ('2494 2126 12\n'
                     '2494 2125 12\n'
                     '2493 2126 12\n'
                     '2493 2125 12\n'
                     '1944 1569 12\n'
                     '1943 1569 12\n'
                     '2086 1974 12\n'
                     '2086 1973 12\n'
                     '2085 1974 12\n'
                     '2085 1973 12\n')

        shape_json = load_geojson('./roi_triple.geojson')
        num_tiles = process_geojson(shape_json, max_zoom=12,
                                    tile_format='google',
                                    output_format='{x} {y} {z}')
        self.assertEqual(num_tiles, 10)
        self.assertEqual(mock_stdout.getvalue(), tile_inds)



"""
test_utils.py

@author: DevelopmentSeed

Test all utilities.
"""
import os
import unittest
from pygeotile.tile import Tile

from geodex.utils import _get_quadrant_tiles, format_tile
testpath = os.path.dirname(__file__)


class TestUtils(unittest.TestCase):
    """Test utility functions."""

    def test_get_quadrant_tiles(self):
        """Test ability to get tile children on level of increased zoom."""
        tile = Tile.from_google(0, 0, 0)
        child_tiles = _get_quadrant_tiles(tile)
        self.assertEqual(child_tiles, [Tile.from_google(0, 0, 1),
                                       Tile.from_google(0, 1, 1),
                                       Tile.from_google(1, 0, 1),
                                       Tile.from_google(1, 1, 1)])


    def test_string_output(self):
        """Test that tile details are correctly printed."""

        tile = Tile.from_google(3, 1, 2)

        self.assertEqual(format_tile(tile, 'google'), '3 1 2')
        self.assertEqual(format_tile(tile, 'tms'), '3 2 2')
        self.assertEqual(format_tile(tile, 'quad_tree'), '13')

        self.assertEqual(format_tile(tile, 'google', '{z}-{x}-{y}'), '2-3-1')
        self.assertEqual(format_tile(tile, 'tms', '{z}-{x}-{y}'), '2-3-2')


    def test_overlap(self):
        """Test that overlap between two shapes is correct."""
        pass



"""
main.py

@author: DevelopmentSeed

Core command line functionality for geodex
"""

import sys
import argparse
import logging
from datetime import datetime as dt

from geodex.version import __version__
from geodex.search import process_geojson
from geodex.utils import load_geojson


logger = logging.getLogger(__name__)
max_zoom = 22  # Highest zoom level allowed
tile_format_options = ['google', 'quad_tree', 'tms']


def parse_args(args):
    """Parse command line arguments."""

    desc = 'geodex (v%s)' % __version__
    #dhf = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(description=desc)

    #pparser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--version', help='Print version and exit',
                        action='version', version=__version__)
    parser.add_argument('--log', default=2, type=int, choices=range(0, 6),
                        help='0:all, 1:debug, 2:info, 3:warning, 4:error, 5:critical')

    parser.add_argument('geojson', type=str,
                        help='File path to geojson file containing at least one geospatial boundary.')
    parser.add_argument('zoom', type=int, choices=range(0, max_zoom + 1),
                        help='Zoom level specifying spatial resolution of returned tiles. Must be integer 0-22.',
                        metavar='zoom')
    parser.add_argument('tile-format', type=str, default='google', nargs='?',
                        choices=tile_format_options,
                        help="Specify the tile index format. Default: 'google'")
    parser.add_argument('--output-format', '-o', default='{x} {y} {z}',
                        type=str, help=("Output tile format string."
                                        "Only valid for tile-formats `google` or `tms`."
                                        "Default: '{x} {y} {z}'. Example: '{z}-{x}-{y}'"))

    # turn Namespace into dictionary
    parsed_args = vars(parser.parse_args(args))

    return parsed_args


def cli():
    """Validate input and execute command"""
    args = parse_args(sys.argv[1:])
    logger.setLevel(args.pop('log') * 10)
    #cmd = args.pop('command')

    geojson_fpath = args.get('geojson')
    zoom = args.get('zoom')
    tile_format = args.get('tile-format')
    output_format = args.get('output_format')

    st_dt = dt.now()  # Start time
    logger.info('Elapsed time: %s', st_dt)

    ###################################
    # Run tile search
    ###################################
    # Get all polygon bounds from geojson file
    geojson_bounds = load_geojson(geojson_fpath)

    # Process all polygons
    process_geojson(geojson_bounds, zoom, tile_format, output_format)

    #########################################
    # Print some final details
    #########################################
    delta = dt.now() - st_dt
    logger.info('Elapsed time: %s', delta)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
"""Script for setting up module."""
import io
from os import path
from imp import load_source
from setuptools import setup, find_packages

__version__ = load_source('geodex.version', 'geodex/version.py').__version__

here = path.abspath(path.dirname(__file__))

# get the dependencies and installs
with io.open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '')
                    for x in all_reqs if 'git+' not in x]

setup(name='geodex',
      author='Development Seed',
      author_email='',
      version=__version__,
      description=('A tool to find all geospatial tile indices overlapping an '
                   'arbitrary boundary at an arbitrary zoom.'),
      url='https://github.com/developmentseed/geodex',
      license='MIT',
      classifiers=['Intended Audience :: Developers',
                   'License :: Freeware',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6'],
      keywords='',
      entry_points={'console_scripts': ['geodex=geodex.main:cli']},
      packages=find_packages(exclude=['docs', 'tests*']),
      include_package_data=True,
      install_requires=install_requires,
      dependency_links=dependency_links)



"""
utils.py

@author: developmentseed

Functions used to generate a list of tiles via recursion
"""
from os import path as op

import json
from shapely.geometry import Polygon
from pygeotile.tile import Tile


def _get_quadrant_tiles(tile):
    """Return indicies of tiles at one higher zoom (in google tiling scheme)"""
    ul = (tile.google[0] * 2, tile.google[1] * 2)

    return [Tile.from_google(ul[0], ul[1], tile.zoom + 1),           # UL
            Tile.from_google(ul[0], ul[1] + 1, tile.zoom + 1),       # LL
            Tile.from_google(ul[0] + 1, ul[1], tile.zoom + 1),       # UR
            Tile.from_google(ul[0] + 1, ul[1] + 1, tile.zoom + 1)]   # LR


def _calc_overlap(geom1, geom2):
    """Return area overlap"""

    return geom1.intersection(geom2).area


def load_geojson(geojson_fpath):
    """Load geojson and return all contained polygons.

    Parameters:
    ----------
    geojson_fpath: str
        Filepath of to geojson containing boundaries.

    Returns:
    -------
    bounds: list
        List of geometries read from geojson file."""

    if not op.exists(geojson_fpath):
        raise FileNotFoundError('{} does not exist'.format(geojson_fpath))
    if not op.splitext(geojson_fpath) not in ['.geojson', '.json']:
        raise ValueError('{} should be a .geojson or .json file'.format(geojson_fpath))

    bounds = None
    with open(geojson_fpath, 'r') as geojson_f:
        raw_json = json.loads(geojson_f.read())
        features = raw_json['features']

        bounds = [feat['geometry'] for feat in features
                  if feat['geometry']['type'] in ['Polygon', 'MultiPolygon']]
    return bounds


def format_tile(tile, tile_format, format_str='{x} {y} {z}'):
    """Convert tile to necessary format.

    Parameters
    ----------
    tile: pygeotile.tile.Tile
        Tile object to be formatted.
    tile_format: str
        Desired tile format. `google`, `tms`, or `quad_tree`
    format_str: str
        String to guide formatting. Only used for `google` or `tms`
        (as quad_tree is one value).
        Default: "{x} {y} {z}". Example: "{z}-{x}-{y}"
    """

    if tile_format == 'google':
        td = {key: val for key, val
              in zip(['x', 'y', 'z'], list(tile.google) + [tile.zoom])}
        return format_str.format(**td)

    elif tile_format == 'tms':
        td = {key: val for key, val
              in zip(['x', 'y', 'z'], list(tile.tms) + [tile.zoom])}
        return format_str.format(**td)

    elif tile_format == 'quad_tree':
        return tile.quad_tree

    else:
        raise ValueError('`tile_format`: {} not recognized'.format(tile_format))


def get_overlap_child_tiles(tile, roi_geom, completely_contained=False):
    """Find all children tiles that overlap a boundary

    Parameters
    ----------
    tile: pygeotile.tile.Tile
        Tile that is checked for overlap with `roi_geom`.
    roi_geom: shapely.geometry.shape
        Boundary of region-of-interest.
    completely_contained: bool
        Whether or not a tile is completely contained in the boundary.
        If a tile is found to have 100% overlap with boundary, set to `True`
        and algorithm can avoid calculating overlap for all future child tiles.
        Default False.

    Returns:
    -------
    return_tiles: list of pygeotile.tile.Tile, bool
        Tiles that are children of `tile` and overlap the boundary

    """

    return_tiles = []
    quad_tiles = _get_quadrant_tiles(tile)  # Compute four contained tiles

    # If sub-tiles are completely contained within boundary, no need to compute overlap
    if completely_contained:
        return [[qt, True] for qt in quad_tiles]

    # For each tile, compute overlap with ROI boundary
    for qt in quad_tiles:
        ll, ur = qt.bounds  # Get lower-left and upper-right points
        tile_pts = ((ll[1], ll[0]), (ur[1], ll[0]),
                    (ur[1], ur[0]), (ll[1], ur[0]))
        tile_polygon = Polygon(tile_pts)

        # Calculate overlap of tile with ROI
        overlap_area = _calc_overlap(roi_geom, tile_polygon)

        # If 100% overlap, indicate this to avoid checking overlap in future
        if overlap_area == tile_polygon.area:
            return_tiles.append([qt, True])
        elif overlap_area > 0:
            return_tiles.append([qt, False])

    return return_tiles
